%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Graph colouring code as described in "Backjumping is Exception Handling"
%using exception handling for backjumping.
%
%SICStus version
%
%@authors E. Robbins, A. King, J. M. Howe
%@date 18/3/20
%
%problem(8, Vars, Values, Cs) described the example in the paper
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



:- use_module(library(terms)).
:- use_module(library(lists)).

main(N, Cs) :-
        bb_put(conflicts, []),
	problem(N, Vars, Values, Cs),
	colour(Vars, Values, Cs).

colour(Vars, Values, Cs) :-
	problem_setup(Vars, Values, Cs, Pairs, AllIds),
	search(Pairs, [], AllIds).

problem_setup(Vars, Values, Cs, Pairs, AllIds) :-
	pairs(Vars, 1, Values, Pairs, AllIds),
	setup_checks(Cs).

pairs([], _Id, _Values, [], []).
pairs([Var | Vars], Id, Values, [Var-Values | Pairs], [Id | Ids]) :-
	Var = _-Id,
	NextId is Id + 1,
	pairs(Vars, NextId, Values, Pairs, Ids).

setup_checks([]).
setup_checks([C | Cs]) :-
	post(C),
	setup_checks(Cs).

post(X-XId \= Y-YId) :-
	suspend(X, Y, XId, YId).

:- block suspend(-, ?, ?, ?), suspend(?, -, ?, ?).
suspend(X, Y, XId, YId) :-
		inconsistent(X, Y)
	->
		MaxId is max(XId, YId),
                MinId is min(XId, YId),
                bb_get(conflicts, Cs),
                bb_put(conflicts, [[MinId, MaxId] | Cs])
	;
		true.

inconsistent(X, Y) :- X == Y.

search([], _, AllIds):- 
	succeed(AllIds).
search([(Var-Id)-[Value | RestValues] | Pairs], ConflictIds, AllIds) :-
	catch(
			(
				%format("Id = ~w, Value = ~w, ConflictIds = ~w~n", [Id, Value, ConflictIds]),
                                write(bind-Id-Value),nl,
				bind(Var,Value),
				search(Pairs, [], AllIds)
			),
			ball(Id, CIds),
			(
			        update_conflict(RestValues, CIds, ConflictIds, Id, NewConflictIds),
				format("Id = ~w, Cids = ~w, NewConflictIds = ~w~n", [Id, CIds, NewConflictIds]),
				search([(Var-Id)-RestValues | Pairs], NewConflictIds, AllIds)
			)
	     ).

bind(Var, Value) :- 
        Var = Value, 
        bb_get(conflicts, Cs),
write(conflicts-Cs),nl,
        analyse_all_conflicts(Cs,MaxId,CIds),
write(MaxId-CIds),nl,
        bb_put(conflicts, []),
        (
          nonvar(MaxId) 
        -> 
	  throw(ball(MaxId,CIds))
        ;
          true
        ).


analyse_all_conflicts([],_,_).
analyse_all_conflicts([[MinId,MaxId]|Cs],MaxId,[LowId,MaxId]):-
        find_lowest(Cs, MinId, LowId).

find_lowest([], LowId, LowId).
find_lowest([[MinC, _MaxC] | Cs], MinId, LowId):-
        (
          MinC < MinId
        ->
          find_lowest(Cs, MinC, LowId)
        ;
          find_lowest(Cs, MinId, LowId)
        ). 


update_conflict([], CIds, ConflictIds, Id, NewConflictIds):-
	merge(CIds, ConflictIds, NewConflictIds),
	delete(NewConflictIds, Id, RestConflictIds),
	max_member(MaxId, RestConflictIds),
	%format("backjump = ~w~n", [HighId]),
	throw(ball(MaxId, RestConflictIds)).
update_conflict([_Value | _RestValues], CIds, ConflictIds, _Id, NewConflictIds):-
        %format("backtrack~n",[]),
	merge(CIds, ConflictIds, NewConflictIds).

succeed(_). %only this case required if first solution is enough
succeed(AllIds):-
	max_member(MaxId, AllIds),
	throw(ball(MaxId, AllIds)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Sample problems
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


problem(1, Vars, Values, Cs) :-
	Vars = [One, Two, Three, Four, Five, Six, Seven, Eight, Nine, Ten],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Five, 
	      One \= Seven, 
	      Two \= Three, 
	      Two \= Eight,
	      Three \= Four, 
	      Three \= Nine,
	      Four \= Five, 
	      Four \= Ten,
	      Five \= Six,
	      Six \= Eight, 
	      Six \= Nine,
	      Seven \= Nine, 
	      Seven \= Ten,
	      Eight \= Ten].

problem(2, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Three, 
	      One \= Four, 
	      One \= Five, 
	      Two \= Three, 
	      Two \= Five, 
	      Three \= Five, 
	      Four \= Six].

problem(3, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Three, 
	      One \= Four, 
	      One \= Five, 
	      Two \= Five, 
	      Three \= Five, 
	      Four \= Six].

problem(4, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Three, 
	      One \= Four, 
	      One \= Five, 
	      Two \= Three, 
	      Two \= Five, 
	      Three \= Five].

problem(5, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Three, 
	      One \= Four, 
	      Two \= Three, 
	      Two \= Four, 
	      Three \= Four].

problem(6, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four],
	Values = [r, g, b],
	Cs = [One \= Two, 
	      One \= Three, 
	      One \= Four, 
	      Two \= Three, 
	      Three \= Four].

problem(7, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five],
	Values = [r, g],
	Cs = [One \= Three, 
	      Two \= Four, 
	      Two \= Five, 
	      Three \= Five].

problem(8, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g],
	Cs = [One \= Three, 
	      Two \= Five, 
	      Two \= Six, 
	      Three \= Four, 
	      Three \= Six].

problem(9, Vars, Values, Cs):-
	Vars = [One, Two, TwoPointFive, Three, Four, Five, Six],
	Values = [r, g],
	Cs = [One \= Three, 
	      Two \= Five, 
	      Two \= Six, 
              TwoPointFive \= Six,
	      Three \= Four, 
	      Three \= Six].


problem(10, Vars, Values, Cs):-
	Vars = [One, Two, Three, ThreePointFive, Four, Five, Six],
	Values = [r, g],
	Cs = [One \= Three, 
	      Two \= Five, 
	      Two \= Six, 
	      Three \= Four, 
	      Three \= Six,
              ThreePointFive \= Six].

%library code
merge([], L2, L2):-!.
merge(L1, [], L1):-!.
merge([N | L1], [M | L2], Out):-
    N < M,!,
    merge(L1, [M | L2], Merge),
    Out = [N | Merge].
merge([N | L1], [M | L2], Out):-
    N == M,!,
    merge(L1, [M | L2], Out).
merge([N | L1], [M | L2], Out):-
    M <N,
    merge([N | L1], L2, Merge),
    Out = [M | Merge].
