%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Graph colouring code using exception handling for backjumping.  This 
%version does not use delay declarations for coroutining constraints,
%rather it checks a full list of constraints
%
%SICStus and SWI version
%
%@authors E. Robbins, A. King, J. M. Howe
%@date 11/3/20
%
%problem(8, Vars, Values, Cs) described the example in the paper
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(terms)).
:- use_module(library(lists)).

main(N, Cs) :-
	problem(N, Vars, Values, Cs),
	colour(Vars, Values, Cs).

colour(Vars, Values, Cs) :-
	problem_setup(Vars, Values, Cs, Pairs, AllIds),
	search(Pairs, [], Cs, AllIds).

problem_setup(Vars, Values, _Cs, Pairs, AllIds) :-
	pairs(Vars, 1, Values, Pairs, AllIds).

pairs([], _Id, _Values, [], []).
pairs([Var | Vars], Id, Values, [Var-Values | Pairs], [Id | Ids]) :-
	Var = _-Id,
	NextId is Id + 1,
	pairs(Vars, NextId, Values, Pairs, Ids).

check_consistency([]).
check_consistency([Con | Cons]):-
        check_constraint(Con),
        check_consistency(Cons).

check_constraint((X-XId \= Y-YId)):-
        ground(X), ground(Y),!,
        (
		inconsistent(X, Y)
	->
		MaxId is max(XId, YId),
                MinId is min(XId, YId),
		throw(ball(MaxId, [MinId, MaxId])) 
	;
		true
        ).
check_constraint(_).

inconsistent(X, Y) :- X == Y.

search([], _, _, AllIds):-
	succeed(AllIds).
search([(Var-Id)-[Value | RestValues] | Pairs], ConflictIds, Cs, AllIds) :-
	catch(
			(
				format("Id = ~w, Value = ~w, ConflictIds = ~w~n", [Id, Value, ConflictIds]),
				bind(Var,Value),
                                check_consistency(Cs),
				search(Pairs, [], Cs, AllIds)
			),
			ball(Id, CIds),
			(
			        update_learning(RestValues, CIds, ConflictIds, Id, NewConflictIds),
				format("Id = ~w, Cids = ~w, NewConflictIds = ~w~n", [Id, CIds, NewConflictIds]),
				search([(Var-Id)-RestValues | Pairs], NewConflictIds, Cs, AllIds)
			)
	     ).

bind(Var, Value) :- Var = Value.

update_learning([], CIds, ConflictIds, Id, NewConflictIds):-
	merge(CIds, ConflictIds, NewConflictIds),
	delete(NewConflictIds, Id, RestConflictIds),
	max_member(MaxId, RestConflictIds),
	format("backjump = ~w~n", [MaxId]),
	throw(ball(MaxId, RestConflictIds)).
update_learning([_Value | _RestValues], CIds, ConflictIds, _Id, NewConflictIds):-
        format("backtrack~n",[]),
	merge(CIds, ConflictIds, NewConflictIds).

succeed(_).
succeed(AllIds):-
	max_member(MaxId, AllIds),
	throw(ball(MaxId, AllIds)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Sample problems
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


problem(1, Vars, Values, Cs) :-
	Vars = [One, Two, Three, Four, Five, Six, Seven, Eight, Nine, Ten],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Five, One \= Seven, 
		  Two \= Three, Two \= Eight,
		  Three \= Four, Three \= Nine,
		  Four \= Five, Four \= Ten,
		  Five \= Six,
		  Six \= Eight, Six \= Nine,
		  Seven \= Nine, Seven \= Ten,
		  Eight \= Ten].

problem(2, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Three, One \= Four, One \= Five, Two \= Three, Two \= Five, Three \= Five, Four \= Six].

problem(3, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Three, One \= Four, One \= Five, Two \= Five, Three \= Five, Four \= Six].

problem(4, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Three, One \= Four, One \= Five, Two \= Three, Two \= Five, Three \= Five].

problem(5, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Three, One \= Four, Two \= Three, Two \= Four, Three \= Four].

problem(6, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four],
	Values = [r, g, b],
	Cs = [One \= Two, One \= Three, One \= Four, Two \= Three, Three \= Four].

problem(7, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five],
	Values = [r, g],
	Cs = [One \= Three, Two \= Four, Two \= Five, Three \= Five].

problem(8, Vars, Values, Cs):-
	Vars = [One, Two, Three, Four, Five, Six],
	Values = [r, g],
	Cs = [One \= Three, Two \= Five, Two \= Six, Three \= Four, Three \= Six].

%library code
merge([], L2, L2):-!.
merge(L1, [], L1):-!.
merge([N | L1], [M | L2], Out):-
    N < M,!,
    merge(L1, [M | L2], Merge),
    Out = [N | Merge].
merge([N | L1], [M | L2], Out):-
    N == M,!,
    merge(L1, [M | L2], Out).
merge([N | L1], [M | L2], Out):-
    M <N,
    merge([N | L1], L2, Merge),
    Out = [M | Merge].
