%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Benchmarking for SAT solvers described in "Backjumping is Exception Handling"
%
%SICStus version
%
%@authors E. Robbins, A. King, J. M. Howe
%@date 30/4/20
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:- use_module(library(lists)).
:- use_module(library(file_systems)).
:- use_module(library(system)).
:- use_module(library(timeout)).
:- use_module(parser).
%:- use_module(cdcl_last_uip).
%:- use_module(cdcl_first_uip).
:- use_module(cdcl_first_uip_inst).
:- use_module(static_var_order).

% Top benchmark directory
%benchmark_dir('benchmarks/uf250.1065.100').
%benchmark_dir('benchmarks/uniform_random_3_sat/sat').
%benchmark_dir('benchmarks/testing').
benchmark_dir('benchmarks/samples').
%benchmark_dir('benchmarks').

% Timeout in seconds
timeout(30).

% K values to use.
% If "none" is in this list, it must come first or the latex table header output will get messed up
%choose one k value for learning [none,1,2,4,8,16,32]). 
ks([8]). 

% Run at most the first X benchmarks from each sub directory
max_files(10). 

% How to sort files. One of size, filename or off
sort_files_setting(filename). 

% Reverse order in which files are sorted
reverse_order :- false.

run_benchmarks :-
	generate_results_filename(Results),
	benchmark_dir(Dir),
	open(Results, write, ResultsStream),
	output_latex_table_header(ResultsStream),
	run_benchmarks_aux(ResultsStream, Dir),
	output_latex_table_footer(ResultsStream),
	close(ResultsStream).

% Run benchmarks in directory Dir, then recurse into sub dirs
run_benchmarks_aux(ResultsStream, Dir) :-
	% Get a list of dimacs files in directory
	findall(FileName, file_member_of_directory(Dir, '*.cnf', _, FileName), Files),
	% Sort files according to options given at top of this file
	sort_files_setting(SortSetting),
	sort_files(SortSetting, Files, SortedFiles),
	% Reverse the file list if set to
	(
		reverse_order
	->
		reverse(SortedFiles, Files1)
	;
		Files1 = SortedFiles
	),
	% Get the first X files from directory
	length(Files1, NumFiles),
	max_files(MaxFilesToRun),
	(
		NumFiles > MaxFilesToRun
	->
		append_length(FileListFinal, _, Files1, MaxFilesToRun)
	;
		FileListFinal = Files1
	),
	% Run the benchmarks
	maplist(benchmark(ResultsStream, Dir), FileListFinal),
	% Recurse over subdirectories
	findall(DirName, directory_member_of_directory(Dir, _, DirName), Directories),
	maplist(run_benchmarks_aux(ResultsStream), Directories).

sort_files(off, Files, Files).
sort_files(filename, Files, SortedFiles) :-
	sort(Files, SortedFiles).
sort_files(size, Files, SortedFiles) :-
	length(Files, NumFiles),
	length(Sizes, NumFiles),
	maplist(file_size, Files, Sizes),
	length(FilesWithSizes, NumFiles),
	maplist(zip, Sizes, Files, FilesWithSizes),
	sort(FilesWithSizes, FilesSortedBySize),
	length(SortedFiles, NumFiles),
	maplist(unzip, FilesSortedBySize, SortedFiles).

file_size(File, Size) :-
	file_property(File, size_in_bytes, Size).

zip(A, B, A-B).
unzip(_A-B, B).

% Run each solver over dimacs file
benchmark(ResultsStream, Directory, File) :-
	format("Name: ~w~n", [File]),
	% Output the file name to latex table (first escape underscores and remove the file extension)
	file_member_of_directory(Directory, BaseName, File),
	escape_underscores(BaseName, EscapedBaseName),
	remove_file_extension(EscapedBaseName, FinalName),
	format(ResultsStream, "~w ", [FinalName]),
	% Set up
        %option 1: just read in the problem
	parser(File, Clauses, Vars),
        %or option 2: static variable ordering
%	parser(File, Clauses, InVars),
%        order_variables(Clauses, InVars, TmpVars),
%        reverse(TmpVars, Vars),!,
	ks(Ks),
	length(Ks, NumRuns),
	length(Data, NumRuns),
	length(Results, NumRuns),
	timeout(TimeOut),
	Timeout is TimeOut * 1000,
	% Run with learning
	maplist(cdcl_benchmark(Clauses, Vars, Timeout), Ks, Data, Results),
	% Run FLOPS (vanilla) solver
	statistics(total_runtime, _),
	time_out(
		   sat(Clauses, Vars, Ass, Sat),
		   Timeout, 
		   Result
		),
	statistics(total_runtime, [_, Time]),
        format("No learning ", []),
        (
             Sat == unsat  
	->
	     format(" ... (unsat)", [])
        ;
             true
        ),
%	statistics(runtime, [_, Time]),
	(
		Result == time_out
	->
		format(" ... timed out~n", [])
	;
		format(" ... ~w ms ~w ass ~n", [Time, Ass])
	),
	% Do some output to table
	output_time(ResultsStream, Time),
	(
                nonvar(Ass) 
        -> 
                output_data(ResultsStream, Ass)
        ; 
                output_data(ResultsStream,'')
        ),
        output_learning(ResultsStream, Data),
	format(ResultsStream, "\\\\~n",[]).

output_learning(_ResultsStream, []).
output_learning(ResultsStream, [D|Ds]):-
        D = [Time | Rest],
	output_time(ResultsStream, Time),
	maplist(output_data(ResultsStream), Rest),
        output_learning(ResultsStream, Ds).

% Output a time to the table or if the result timed out then output e.g. ">60000" for a 60 second timeout
output_time(ResultsStream, Time) :-
	timeout(TimeoutS),
	TimeoutMS is TimeoutS * 1000,
	(
		Time >= TimeoutMS
	->
		format(ResultsStream, "& $>$~w ", [TimeoutMS])
	;
		format(ResultsStream, "& ~w ", [Time])
	).

% Output performance stats
output_data(ResultsStream, Data) :-
        format(ResultsStream, "& ~w ", [Data]).

% Run the CDCL solver with a specific K
cdcl_benchmark(Clauses, Vars, Timeout, K, [Time, Ass, Throws, Jumps], Result) :-
	format("K = ~w", [K]),
	(
		K == none
	->
		ActualK = 0
	;
		ActualK = K
	),
	statistics(total_runtime, _),
%	statistics(runtime, _),
	copy_term((Clauses, Vars), (ClausesCopy, VarsCopy)),
	time_out(
			      	cdcl(ClausesCopy, VarsCopy, ActualK, Result, Ass, Throws, Jumps), 
				Timeout,
				TimedOut
			),
        (
             Result == unsat  
	->
	     format(" ... (unsat)", []),
	     Result = unsat
        ;
             true
        ),				
%	statistics(runtime, [_, Time]),
	statistics(total_runtime, [_, Time]),
	(
		TimedOut == time_out,
		Result = time_out,
		format(" ... timed out~n", [])
	;
		TimedOut \== time_out,
		(
			Result == unsat
		; 
			Result \== unsat, Result = sat
		),
		format(" ... ~w ms, ~w ass, ~w throws, ~w jumps~n", [Time, Ass, Throws, Jumps])
	).

output_latex_table_header(ResultsStream) :-
	format(ResultsStream, "\\begin{table}[t!]~n", []),
	format(ResultsStream, "\\centering~n", []),
	format(ResultsStream, "\\caption{Performance of learning versus no learning}~n", []),
	format(ResultsStream, "\\begin{tabular}{@{}l|@{}r @{}r | @{}r @{}r @{}r @{}r }~n",[]),
        format(ResultsStream, "& \\multicolumn{2}{c|}{no learning} & \\multicolumn{4}{c}{$8$-learning} \\\\~n",[]),
        format(ResultsStream, "\\cline{1-7}~n", []),
        format(ResultsStream, "benchmark & time & assign & time & assign & throws & jumps \\\\~n", []),
        format(ResultsStream, "\\cline{1-7}~n", []).
	
output_column_headers_row_1(ResultsStream, []) :-
	format(ResultsStream, "}~n", []).
output_column_headers_row_1(ResultsStream, [none| Ks]) :-
	format(ResultsStream, "r|@{}", []),
	output_column_headers_row_1(ResultsStream, Ks).
output_column_headers_row_1(ResultsStream, [_ | Ks]) :-
	format(ResultsStream, "r@{}", []),
	output_column_headers_row_1(ResultsStream, Ks).

output_latex_table_footer(ResultsStream) :-
	format(ResultsStream, "\\end{tabular}~n", []),
	format(ResultsStream, "\\label{tbl:results}~n", []),
	format(ResultsStream, "\\end{table}~n", []).

% Generate a file name with a time stamp
generate_results_filename(Name) :-
	name('results_', Name1),
	name('_', Underscore),
	name('.tex', DotTex),
	datime(datime(Year,Month,Day,Hour,Min,_Sec)),
	number_codes(Year, YearCodes),
	number_codes(Month, MonthCodes),
	number_codes(Day, DayCodes),
	number_codes(Hour, HourCodes),
	number_codes(Min, MinCodes),
	append(Name1, YearCodes, Name2),
	append(Name2, Underscore, Name3),
	append(Name3, MonthCodes, Name4),
	append(Name4, Underscore, Name5),
	append(Name5, DayCodes, Name6),
	append(Name6, Underscore, Name7),
	append(Name7, HourCodes, Name8),
	append(Name8, Underscore, Name9),
	append(Name9, MinCodes, Name10),
	append(Name10, DotTex, Name11),
	name(Name, Name11).

escape_underscores(String, EscapedString) :-
	name(String, StringCodes),
	escape_underscores_aux(StringCodes, EscapedStringCodes),
	name(EscapedString, EscapedStringCodes).

%92 = backslash
%95 = underscore
escape_underscores_aux([], []).
escape_underscores_aux([Char | Chars], [92, 95 | NextChars]) :-
	Char == 95,
	!,
	escape_underscores_aux(Chars, NextChars).
escape_underscores_aux([Char | Chars], [Char | NextChars]) :-
	escape_underscores_aux(Chars, NextChars).

remove_file_extension(FileName, FileNameSansExt) :-
	name(FileName, FileNameCodes),
	length(FileNameCodes, NameLength),
	NewNameLength is NameLength - 4,
	append_length(FileNameSansExtCodes, _, FileNameCodes, NewNameLength),
	name(FileNameSansExt, FileNameSansExtCodes).

% FLOPS SAT Solver below (no learning)

sat(Clauses, Vars, Ct, sat) :- 
        bb_put(ct,0),
        sat_aux(Clauses, Vars),!,
        bb_get(ct,Ct).%,!,
sat(_Clauses, _Vars, Ct, unsat) :-         
        bb_get(ct,Ct).


sat_aux(Clauses, Vars) :- 
	problem_setup(Clauses), search(Vars).


search([]).
search([Var | Vars]) :-
        (bind(Var,true);bind(Var,false)),
%        (Var = true; Var = false),
        search(Vars).


bind(Var,Pol):-
        (
           var(Var)
        ->
           bb_get(ct,Ct),NewCt is Ct+1,bb_put(ct,NewCt),
           Var=Pol
        ;
           Var=Pol
        ).

assign(true).
assign(false).

problem_setup([]).
problem_setup([Clause | Clauses]) :-
	clause_setup(Clause),
	problem_setup(Clauses).

clause_setup([Pol-Var | Pairs]) :-
	set_watch(Pairs, Var, Pol).

set_watch([], Var, Pol) :-
	Var = Pol.
set_watch([Pol2-Var2 | Pairs], Var1, Pol1):-
	watch(Var1, Pol1, Var2, Pol2, Pairs).

:- block watch(-, ?, -, ?, ?).
watch(Var1, Pol1, Var2, Pol2, Pairs) :-
	(
		nonvar(Var1) 
	->
		update_watch(Var1, Pol1, Var2, Pol2, Pairs)
	;
		update_watch(Var2, Pol2, Var1, Pol1, Pairs)
	).

update_watch(Var1, Pol1, Var2, Pol2, Pairs) :-
	(
		Var1 == Pol1
	;
		Var1 \== Pol1,
		set_watch(Pairs, Var2, Pol2)
	).

