:- module(cdcl, [dimacs/2, cdcl/3]).
:- use_module(library(lists)).
:- use_module(library(assoc)).
:- use_module(library(between)).
:- use_module(parser).
:- use_module(auxiliaries).
:- use_module(static_var_order).

:- dynamic learnt/1.

% Smaller optimisations not used here but used in benchmarking:
% (1) Pass back highest level so clock always goes forward; this speeds up add_learnt_clauses
% (2) Use arg instead of assoc for variable IDs to simulate an array; this speeds up unground

% parse a dimacs file to get a clause and then run the solver with k learning of K
dimacs(File, K) :-
	parser(File, Clauses, Vars), 
	cdcl(Clauses, Vars, K).

% set up the search and kick it off
cdcl(Clauses, Vars, K) :-
	put_learnt([]),				% subtle state issue here for assert/retract
	ids(Vars, Ids, IdMap),
	vars(Vars, Ids),
	watch_clauses(Clauses),	
	(
		K == none
	->
		search_setup(Vars, IdMap, 1)
	;
		search_k_learning_setup(Vars, IdMap, 1, K)
	).
%cdcl(_,_,_):-get_learnt(Ls), length(Ls, L), write(len-L),nl,fail.

%ids([Var | Vars], Ids, Id_Map) :-
%	length(Vars, Length),
%	numlist(0, Length, Ids),
%	maplist(zip, Ids, [Var | Vars], IdsVars),
%	list_to_assoc(IdsVars, Id_Map).

ids(Vars, Ids, Id_Map) :-
	length(Vars, Length),
	numlist(1, Length, Ids),
	(true ->
		create_map(Vars, Arrays),
		Id_Map =.. [array | Arrays]
	;
		maplist(zip, Ids, Vars, IdsVars),
		list_to_assoc(IdsVars, Id_Map)
	).

create_map([], []).
create_map([Var | Vars], [Array | Arrays]) :-
	length([Var | Vars], Size),
	Arity is min(Size, 128),
	length(Prefix, Arity),
	Array =.. [array | Prefix],
	append(Prefix, Residual, [Var | Vars]),
	create_map(Residual, Arrays).

zip(X, Y , X-Y).

vars([], []).
vars([_-imp(_Level, Id, _Pol, _SubWhys) | Vars], [Id | Ids]) :-
	vars(Vars, Ids).

get_learnt(Learnt) :- bb_get(learnt, Learnt).
put_learnt(Learnt) :- bb_put(learnt, Learnt).
%get_learnt(Learnt) :- retract(learnt(Learnt)).%bb_get(learnt, Learnt).
%put_learnt(Learnt) :- assert(learnt(Learnt)).%bb_put(learnt, Learnt).
% end set up code

watch_clauses([]).
watch_clauses([Clause | Clauses]) :-
	watch_clause(Clause),
	watch_clauses(Clauses).

watch_clause([Pol-VarWhy | Pairs]) :-
	set_watch(Pairs, VarWhy, Pol, []).

set_watch([], Var-Why, Pol, Whys) :-
	(
		(nonvar(Var), Var \== Pol)
	->
		conflict([Why | Whys])
	;
		unit(Var-Why, Pol, Whys)
	).
set_watch([Pol2-(Var2-Why2) | Pairs], Var1-Why1, Pol1, Whys) :-
	watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys).

:- block watch(-, ?, ?, -, ?, ?, ?, ?).
watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys) :-
	(
		nonvar(Var1)
	->
		update_watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys)
	;
		update_watch(Var2, Why2, Pol2, Var1, Why1, Pol1, Pairs, Whys)
	).

update_watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys) :-
	(
		Var1 == Pol1
	->
		true
	;
		set_watch(Pairs, Var2-Why2, Pol2, [Why1 | Whys])
	).

unit(Var-Why, Pol, Whys) :-
	(
		nonvar(Var)
	-> 
		true
	;
		max_member(imp(MaxLevel, _, _, _), [imp(0-0, _, _, _) | Whys]),	% NOTE: sublevels to support first UIP
		increment_sublevel(MaxLevel, MaxLevel1),			% NOTE: sublevels to support first UIP
		Why = imp(MaxLevel1, _Id, Pol, Whys),
		Var = Pol
	).

% NOTE: sublevels to support first UIP
increment_sublevel(Level-SubLevel, Level-SubLevel1) :-
	SubLevel1 is SubLevel + 1.

search_setup(Vars, IdMap, Level):-
        catch(
                 (
                     search(Vars, IdMap, Level)
                 ),
                 ball(0, Clause),
                 (
                     update_learning(1,Clause,0,IdMap),
                     search_setup(Vars, IdMap, Level)
                 )
             ).

search([], _, _).
search([Var-Why | VarWhys], IdMap, Level) :-
	NextLevel is Level + 1,
	catch(
			(
				bind(Var-Why, Level),
				search(VarWhys, IdMap, NextLevel) 
			),
			ball(Level, Clause),
			(
                                update_learning(NextLevel, Clause, Level, IdMap),
				search([Var-Why | VarWhys], IdMap, NextLevel)
			)
	     ).

% NOTE: sublevels to support first UIP
bind(Var-Why, Level):-
        (
            var(Var)
        ->
            bind_aux(Var-Why, Level)
	;
            true
        ).

bind_aux(true-imp(Level-0, _, true, []), Level).%:- (V=true; trace, fail).
%bind_aux(false-imp(Level-0, _, false, []), Level).

update_learning(NextLevel, Clause, Level, IdMap):-
	get_learnt(Learnt),
	Learnt1 = [NextLevel-Clause | Learnt],
	add_learnt_clauses(Learnt1, Level, IdMap),
	put_learnt(Learnt1).


conflict(Whys) :-
	analyse_conflict(Whys, BackjumpLevel, Clause),
        throw(ball(BackjumpLevel, Clause)).

% start first UIP
analyse_conflict(Whys, BackjumpLevel, Clause) :-
	max_member(imp(MaxLevel, _, _, _), [imp(0-0, _, _, _) | Whys]),
	construct_clause(Whys, MaxLevel, Lits, Levels),
	remove_dups(Lits, Clause),
	sort(Levels, SortedLevels),
	(
		SortedLevels = [BackjumpLevel]
	->
		BackjumpLevel \== 0
	;
		reverse(SortedLevels, [_, BackjumpLevel | _])
%		reverse(SortedLevels, [BackjumpLevel | _])
	).


construct_clause(Whys, MaxLevel-_, [NegPol-Id | Lits], [Level | Levels]) :-
	first_uip(Whys, imp(MaxLevel-0, _, _, _), Frontier),
	last(Rest, UIP, Frontier),
	UIP = imp(Level-_, Id, Pol, _),
	negate(Pol, NegPol),
	remove_dups(Rest, CleanRest),
	last_uip(CleanRest, Lits, Levels).

first_uip(Whys, PivotWhy, Frontier) :-
	partition_whys(Whys, PivotWhy, [HighWhy | HighWhys], LowWhys),
	append(LowWhys, RestFrontier, Frontier),
	highest_why(HighWhys, HighWhy, HighestWhy, ResidualWhys),
	first_uip(ResidualWhys, PivotWhy, HighestWhy, RestFrontier).

first_uip([], _PivotWhy, HighestWhy, [HighestWhy]).
first_uip([Why | Whys], PivotWhy, imp(_, _, _, SubWhys), Frontier) :-
	partition_whys(SubWhys, PivotWhy, HighSubWhys, LowSubWhys),
	append(HighSubWhys, Whys, NextWhys),
	append(LowSubWhys, RestFrontier, Frontier),
	highest_why(NextWhys, Why, HighestWhy, ResidualWhys),
	first_uip(ResidualWhys, PivotWhy, HighestWhy, RestFrontier).

partition_whys([], _PivotWhy, [], []).
partition_whys([Why | Whys], PivotWhy, HighWhys, LowWhys) :-
	(
		Why @< PivotWhy
	->
		LowWhys = [Why | RestLowWhys],
		partition_whys(Whys, PivotWhy, HighWhys, RestLowWhys)
	;
		HighWhys = [Why | RestHighWhys],
		partition_whys(Whys, PivotWhy, RestHighWhys, LowWhys)
	).

highest_why([], HighWhy, HighWhy, []).
highest_why([Why | Whys], HighWhy, HighestWhy, LowWhys) :-
	compare(Op, HighWhy, Why),
	highest_why(Op, HighWhy, HighestWhy, LowWhys, Why, Whys).

highest_why(<, HighWhy, HighestWhy, [HighWhy | LowWhys], Why, Whys) :-
	highest_why(Whys, Why, HighestWhy, LowWhys).
highest_why(=, HighWhy, HighestWhy, LowWhys, _Why, Whys) :-
	highest_why(Whys, HighWhy, HighestWhy, LowWhys).
highest_why(>, HighWhy, HighestWhy, [Why | LowWhys], Why, Whys) :-
	highest_why(Whys, HighWhy, HighestWhy, LowWhys).

simplify_frontier([], Acc, Acc).
simplify_frontier([Why | Whys], Acc, Literals) :-
	(
		redundant(Why, Whys)
	->
		simplify_frontier(Whys, Acc, Literals)
	;
		simplify_frontier(Whys, [Why | Acc], Literals)
	).

redundant(imp(_, _, _, [SubWhy | SubWhys]), Whys) :-
	member_decreasing_set(Whys, SubWhy),
	redundants(SubWhys, Whys).

redundants([], _Whys).
redundants([SubWhy | SubWhys], Whys) :-
	(
		member_decreasing_set(Whys, SubWhy)
	->
		true
	;
		redundant(SubWhy, Whys)
	),
	redundants(SubWhys, Whys).

member_decreasing_set([X | Xs], Y) :-
	compare(Op, X, Y),
	member_decreasing_set(Op, Y, Xs).

member_decreasing_set(=, _Y, _Xs).
member_decreasing_set(>, Y, [X | Xs]) :-
	compare(Op, X, Y),
	member_decreasing_set(Op, Y, Xs).

last_uip([], [], []).
last_uip([imp(Level-_, Id, Pol, SubWhys) | Whys], Lits, Levels) :-
	(
		SubWhys == []
	->
		negate(Pol, NegPol),
		Lits = [NegPol-Id | MoreLits],
		Levels = [Level | MoreLevels],
		last_uip(Whys, MoreLits, MoreLevels)
	;
		append(SubWhys, Whys, AllWhys),
		last_uip(AllWhys, Lits, Levels)
	).

negate(true, false).
negate(false, true).
% end first UIP

% code to add learnt clauses 
add_learnt_clauses([], _, _).
add_learnt_clauses([Level-Clause | Clauses], DecisionLevel, IdMap) :-
	(
		Level < DecisionLevel
	->
		true
	;
		unground(Clause, IdMap, NewClause),
		watch_clause(NewClause)
	),
	add_learnt_clauses(Clauses, DecisionLevel, IdMap).

unground([], _, []).
unground([Pol-Id | Labels], IdMap, [Pol-VarImp | Literals]) :-
	(functor(IdMap, array, _) ->
		Id1 is Id - 1,
		OuterIndex is div(Id1, 128) + 1,
		InnerIndex is mod(Id1, 128) + 1,
		arg(OuterIndex, IdMap, Array),
		arg(InnerIndex, Array, VarImp)
	;
		get_assoc(Id, IdMap, VarImp)
	),
	unground(Labels, IdMap, Literals).
% end add learnt clauses code



% k learning version of search

search_k_learning_setup(Vars, IdMap, Level, K):-
        catch(
                 (
                     search_k_learning(Vars, IdMap, Level, K)
                 ),
                 ball(0, Clause),
                 (
                     update_k_learning(1, Clause, 0, IdMap, K),
                     search_k_learning_setup(Vars, IdMap, Level,K)
                 )
             ).

search_k_learning([], _, _, _).
search_k_learning([Var-Why | VarWhys], IdMap, Level, K) :-
	NextLevel is Level + 1,
	catch(	  (
			bind(Var-Why, Level),
			search_k_learning(VarWhys, IdMap, NextLevel, K) 
		  ),
		  ball(Level, Clause),
		  (
                        update_k_learning(NextLevel, Clause, Level, IdMap, K),
			search_k_learning([Var-Why | VarWhys], IdMap, NextLevel, K)
		  )
	     ).

% end k learning

update_k_learning(NextLevel, Clause, Level, IdMap, K):-
	get_learnt(Learnt),
	Learnt1 = [NextLevel-Clause | Learnt],
	add_learnt_clauses(Learnt1, Level, IdMap),
        length(Clause, ClauseLength),
	(
	     ClauseLength =< K
	->
	     (
	         new_learnt_clauses(Learnt, Clause, NewLearnt),
                 put_learnt([NextLevel-Clause | NewLearnt])
             )
	;
	     true % put_learnt(Learnt) 
	).

new_learnt_clauses([], _, []).
new_learnt_clauses([Level-LearntClause | LearntClauses], Clause, NewLearnt):-
        (
            	%dominated(Clause, LearntClause)
            	ordered_subset(Clause, LearntClause)
        -> 
            	new_learnt_clauses(LearntClauses, Clause, NewLearnt)
        ;
            	new_learnt_clauses(LearntClauses, Clause, RestLearnt),
            	NewLearnt = [Level-LearntClause | RestLearnt]
        ).

dominated([], _).
dominated([C|Cs], LearntClause):- 
        dominated(Cs, LearntClause),
        my_memberchk(LearntClause,C).

ordered_subset([], _Ys).
ordered_subset([X | Xs], [Y | Ys]) :-
	(
		X == Y 
	->
		ordered_subset(Xs, Ys)
	;
		Y @< X, 
		ordered_subset([X | Xs], Ys)
	).

my_memberchk([C | _LCs], C):-!.
my_memberchk([_ | LCs], C):-
        my_memberchk(LCs, C).
