%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Last UIP SAT solver described in "Backjumping is Exception Handling"
%
%SICStus version
%
%@authors E. Robbins, A. King, J. M. Howe
%@date 8/5/19
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module(cdcl, [dimacs/2, cdcl/3]).
:- use_module(library(lists)).
:- use_module(library(assoc)).
:- use_module(library(between)).
:- use_module(parser).
:- use_module(auxiliaries).

:- dynamic learnt/1.
% Smaller optimisations not used here but used in benchmarking:
% (1) Pass back highest level so clock always goes forward; this speeds up add_learnt_clauses
% (2) Use arg instead of assoc for variable IDs to simulate an array; this speeds up unground

% parse a dimacs file to get a clause and then run the solver with k learning of K
dimacs(File, K) :-
	parser(File, Clauses, Vars), 
	cdcl(Clauses, Vars, K).

% set up the search and kick it off
% begin Fig 5 (note conditional for experiments)
cdcl(Clauses, Vars, K) :-
	put_learnt([]),
	ids(Vars, Ids, IdMap),
	vars(Vars, Ids),
	watch_clauses(Clauses),	
	(
		K == 0
	->
                search_setup(Vars, IdMap, 1)
	;
		search_k_learning_setup(Vars, IdMap, 1, K)
	).

ids([Var | Vars], Ids, IdMap) :-
	length(Vars, Length),
	numlist(0, Length, Ids),
	maplist(zip, Ids, [Var | Vars], IdsVars),
	list_to_assoc(IdsVars, IdMap).

zip(X, Y , X-Y).

vars([], []).
vars([_-imp(_Level, Id, _Pol, _SubWhys) | Vars], [Id | Ids]) :-
	vars(Vars, Ids).

%end Fig 5

% included in Fig 4
watch_clauses([]).
watch_clauses([Clause | Clauses]) :-
	watch_clause(Clause),
	watch_clauses(Clauses).
%end included in Fig 4

% start fig 6
watch_clause([Pol-VarWhy | Pairs]) :- set_watch(Pairs, VarWhy, Pol, []).

set_watch([], Var-Why, Pol, Whys) :-
	(
		(nonvar(Var), Var \== Pol)
	->
		conflict([Why | Whys])
	;
		unit(Var-Why, Pol, Whys)
	).
set_watch([Pol2-(Var2-Why2) | Pairs], Var1-Why1, Pol1, Whys) :-
	watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys).

:- block watch(-, ?, ?, -, ?, ?, ?, ?).
watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys) :-
	(
		nonvar(Var1)
	->
		update_watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys)
	;
		update_watch(Var2, Why2, Pol2, Var1, Why1, Pol1, Pairs, Whys)
	).

update_watch(Var1, Why1, Pol1, Var2, Why2, Pol2, Pairs, Whys) :-
	(
		Var1 == Pol1
	->
		true
	;
		set_watch(Pairs, Var2-Why2, Pol2, [Why1 | Whys])
	).

unit(Var-Why, Pol, Whys) :-
	(
		nonvar(Var)
	-> 
		true
	;
		max_member(imp(MaxLevel, _, _, _), [imp(0-0, _, _, _) | Whys]),	% NOTE: sublevels to support first UIP
		increment_sublevel(MaxLevel, MaxLevel1),			% NOTE: sublevels to support first UIP
		Why = imp(MaxLevel1, _Id, Pol, Whys),
		Var = Pol
	).

% NOTE: sublevels to support first UIP
increment_sublevel(Level-SubLevel, Level-SubLevel1) :-
	SubLevel1 is SubLevel + 1.

% end fig 6



% start fig 7

search_setup(Vars, IdMap, Level):-
        catch(
                 (
                     search(Vars, IdMap, Level)
                 ),
                 ball(0, Clause),
                 (
                     update_conflict(1,Clause,0,IdMap),
                     search_setup(Vars, IdMap, Level)
                 )
             ).

search([], _, _).
search([Var-Why | VarWhys], IdMap, Level) :-
	NextLevel is Level + 1,
	catch(
			(
				bind(Var-Why, Level),
				search(VarWhys, IdMap, NextLevel) 
			),
			ball(Level, Clause),
			(
                                update_conflict(NextLevel, Clause, Level, IdMap),
				search([Var-Why | VarWhys], IdMap, NextLevel)
			)
	     ).

% NOTE: sublevels to support first UIP
bind(Var-Why, Level):-
        (
            var(Var)
        ->
            Var-Why = true-imp(Level-0, _, true, [])
	;
            true
        ).

conflict(Whys) :-
	analyse_conflict(Whys, BackjumpLevel, Clause),
        throw(ball(BackjumpLevel, Clause)).

% end Fig 7


% begin Fig 8

update_conflict(NextLevel, Clause, Level, IdMap):-
	get_learnt(Learnt),
	Learnt1 = [NextLevel-Clause | Learnt],
	add_learnt_clauses(Learnt1, Level, IdMap),
	put_learnt(Learnt1).

add_learnt_clauses([], _, _).
add_learnt_clauses([Level-Clause | Clauses], DecisionLevel, IdMap) :-
	(
		Level < DecisionLevel
	->
		true
	;
		unground(Clause, IdMap, NewClause),
		watch_clause(NewClause)
	),
	add_learnt_clauses(Clauses, DecisionLevel, IdMap).

unground([], _, []).
unground([Pol-Id | Labels], IdMap, [Pol-VarImp | Literals]) :-
	get_assoc(Id, IdMap, VarImp),
	unground(Labels, IdMap, Literals).

%get_learnt(Learnt) :- retract(learnt(Learnt)). %bb_get(learnt, Learnt).
%put_learnt(Learnt) :- assert(learnt(Learnt)). %bb_put(learnt, Learnt).
get_learnt(Learnt) :- bb_get(learnt, Learnt).
put_learnt(Learnt) :- bb_put(learnt, Learnt).

% end Fig 8

% begin Fig 10, with last UIP
analyse_conflict(Whys, BackjumpLevel, Clause) :-
        %MaxLevel not require for last UIP
	%max_member(imp(MaxLevel, _, _, _), [imp(0-0, _, _, _) | Whys]),
	construct_clause(Whys, Lits, Levels),
	remove_dups(Lits, Clause),
	sort(Levels, SortedLevels),
	(
		SortedLevels = [BackjumpLevel]
	->
		BackjumpLevel \== 0
	;
		reverse(SortedLevels, [_, BackjumpLevel | _])
	).

construct_clause([], [], []).
construct_clause([imp(Level-_, Id, Pol, SubWhys) | Whys], Lits, Levels) :-
	(
		SubWhys == []
	->
		negate(Pol, NegPol),
		Lits = [NegPol-Id | MoreLits],
		Levels = [Level | MoreLevels],
		construct_clause(Whys, MoreLits, MoreLevels)
	;
		append(SubWhys, Whys, AllWhys),
		construct_clause(AllWhys, Lits, Levels)
	).

negate(true, false).
negate(false, true).
% end Fig 10 last UIP

% k learning version of search

search_k_learning_setup(Vars, IdMap, Level, K):-
        catch(
                 (
                       search_k_learning(Vars, IdMap, Level, K)
                 ),
                 ball(0, Clause),
                 (
                       %update_conflict(1, Clause, 0, IdMap),
                       update_k_learning(1, Clause, 0, IdMap, K),
                       search_k_learning_setup(Vars, IdMap, Level,K)
                 )
             ).

search_k_learning([], _, _, _).
search_k_learning([Var-Why | VarWhys], IdMap, Level, K) :-
	NextLevel is Level + 1,
	catch(	  (
			bind(Var-Why, Level),
			search_k_learning(VarWhys, IdMap, NextLevel, K) 
		  ),
		  ball(Level, Clause),
		  (
                        update_k_learning(NextLevel, Clause, Level, IdMap, K),
			search_k_learning([Var-Why | VarWhys], IdMap, NextLevel, K)
		  )
	     ).

update_k_learning(NextLevel, Clause, Level, IdMap, K):-
	get_learnt(Learnt),
	Learnt1 = [NextLevel-Clause | Learnt],
	add_learnt_clauses(Learnt1, Level, IdMap),
        length(Clause, ClauseLength),
	(
	     ClauseLength =< K
	->
	     put_learnt(Learnt1)
	;
	     true %put_learnt(Learnt1) %required is assert/retract used
	).

% end k learning

